
'===============================================================================
' Autor ............. Jrgen Schlottke, Schnaich-C.-Str. 46, D-W 2200 Elmshorn
' Zweck ............. Demonstration der Schachprogrammierung unter Turbo-Pascal
' Datum ....................... irgendwann 1991, als PD freigegeben am 18.01.93
' Version ................................................. ohne Versionsnummer
' FreeBasic Bearbeitung .................. Roland Chastain, Sarreguemines, 2013
'===============================================================================

const OW = 000 ' Ohnewert
const BW = 002 ' Bauernwert
const LW = 006 ' Laferwert
const SW = 007 ' Springerwert
const TW = 010 ' Turmwert
const DW = 019 ' Damenwert
const KW = 126 ' Knigswert
const AW = 127 ' Ausserhalb

type tListe ' Zugliste
  as integer v(1 to 100) ' von
  as integer n(1 to 100) ' nach
  as integer w(1 to 100) ' wert
end type

type tBrett
  as integer brett(-10 to 109)
  as integer farbe
  as string * 4 rochade
  as integer epfeld(-1 to 1)
  as tListe liste
  as integer anzahl
  as integer material
  declare function Bewertung(as integer) as integer
  declare sub Setzen(as integer, as integer, as integer = DW)
  declare sub Eintragen(as integer, as integer)
  declare sub ZuggeneratorOhne
  declare function Schach as integer
  declare sub RochadeTesten(as integer, as integer, as integer)
  declare sub Zuggenerator
  declare function RekursiveBewertung(as integer, as integer = 1, as integer = 32000) as integer
end type

type tSchachprogramm
  as tBrett grundbrett, spielbrett
  declare sub Grundstellung(as string = "")
  declare function HtmlAnfang(as string = "chess alfonso-x", as string = "32px", as string = "midnightblue") as string
  declare function Brettdarstellung as string
  declare function HtmlEnde as string
  declare function Spielereingabe(as string) as integer
  declare function Zugstring(as integer, as integer) as string
  declare function FeldIstInGrundstellung(byref as tBrett, as integer) as integer
  declare function Besterzug(as integer) as integer
  declare function Computerzug as string
end type

declare sub DeleteChar(as string, as integer)

dim shared as integer bauer   (1 to 4) = {+01,+02,-09,+11}
dim shared as integer springer(1 to 8) = {+12,+21,+19,+08,-12,-21,-19,-08}
dim shared as integer laeufer (1 to 4) = {+11,-11,+09,-09}
dim shared as integer turm    (1 to 4) = {-01,+01,-10,+10}
dim shared as integer koenig  (1 to 8) = {-01,+01,-10,+10,+11,-11,+09,-09}

dim shared as integer halbzugzahl = 0
dim shared as integer rechentiefe = 3
dim shared as integer maxtiefe    = 5

dim shared as tListe liste2
dim shared as integer anzahl2

sub DeleteChar(s as string, index as integer)
  dim aux as string = ""
  for i as integer = 0 to len(s) - 1
    if i <> index - 1 then
      aux += chr(s[i])
    end if
  next i
  s = aux
end sub

function tBrett.Bewertung(f as integer) as integer
  function = material * f
end function

sub tBrett.Setzen(v as integer, n as integer, umwandlung as integer = DW)
  dim char as string * 1 = ""
  dim i as integer
  if brett(v) = TW * farbe then
    select case v
      case 81 : char = "K" 
      case 11 : char = "Q"
      case 88 : char = "k"
      case 18 : char = "q"
    end select
    if (len(char) = 1) then
      i = instr(rochade, char)
      if i > 0 then
        if len(rochade) = 1 then
          rochade = "-"
        else
          DeleteChar(rochade, i)
        end if
      end if
    end if
  end if
  if brett(v) = KW * farbe then
    select case v
      case 51
        i = instr(rochade, "K") : if i > 0 then DeleteChar(rochade, i)
        i = instr(rochade, "Q") : if i > 0 then DeleteChar(rochade, i)
      case 58
        i = instr(rochade, "k") : if i > 0 then DeleteChar(rochade, i)
        i = instr(rochade, "q") : if i > 0 then DeleteChar(rochade, i)
    end select
    if rochade = "" then rochade = "-"
  end if
  if ((n - v) * (n - v) = 4) and (brett(v) = BW * farbe) then
    epfeld(farbe) = v + bauer(1) * farbe
  else
    epfeld(farbe) = 0
  end if
  if (n = epfeld(- 1 * farbe)) and (brett(v) = BW * farbe) then
     epfeld(farbe) = 0
     Setzen(v, (n \ 10) * 10 + v mod 10)
     farbe = - 1 * farbe
     v = (n \ 10) * 10 + v mod 10
  end if
  if (brett(v) = KW * farbe) _
  and ((v = 51) or (v = 58)) _
  and ((n - v) * (n - v) = 400) then
    brett(n) = brett(v)
    brett(v) = 0
    if n \ 10 = 7 then
      v = (v mod 10) + 80
      n = v - 20
    else
      v = (v mod 10) + 10
      n = v + 30
    end if
    brett(n) = brett(v)
    brett(v) = 0
  else
    material = material - brett(n)
    brett(n) = brett(v)
    brett(v) = 0
    if ((n mod 10 = 1) or (n mod 10 = 8)) and (brett(n) = BW * farbe) then
      brett(n) = umwandlung * farbe
      material = material + (umwandlung - BW) * farbe
    end if
  end if
  farbe = -1 * farbe
end sub

sub tBrett.Eintragen(v as integer, n as integer)
  anzahl += 1
  liste.v(anzahl) = v
  liste.n(anzahl) = n
end sub

sub tBrett.ZuggeneratorOhne
  dim as integer v, n
  dim as integer i, j, k, l
  anzahl = 0
  for j = 1 to 8
    for k = 1 to 8
      v = j * 10 + k
      if brett(v) <> OW then
        select case brett(v) * farbe
        case BW
          n = v + bauer(1) * farbe
          if brett(n) = OW then
            Eintragen(v, n)
            if (farbe = 1) and (k = 2) or (farbe = -1) and (k = 7) then
              n = v + bauer(2) * farbe
              if brett(n) = OW then
                Eintragen(v, n)
              end if
            end if
          end if
          for i = 3 to 4
            n = v + bauer(i) * farbe
            if ((-1 * brett(n) * farbe >= BW) _
            and (-1 * brett(n) * farbe <= KW)) _
            or (n = epfeld(-1 * farbe)) then
              Eintragen(v, n)
            end if
          next i
        case SW
          for i = 1 to 8
            n = v + springer(i)
            if (-1 * brett(n) * farbe >= OW) _
            and (-1 * brett(n) * farbe <= KW) then
              Eintragen(v, n)
            end if
          next i
        case LW
          for i = 1 to 4
            n = v
            do
              n += laeufer(i)
              if (-1 * brett(n) * farbe >= OW) _
              and (-1 * brett(n) * farbe <= KW) then
                Eintragen(v, n)
              end if
            loop until brett(n) <> OW
          next i
        case TW
          for i = 1 to 4
            n = v
            do
              n += turm(i)
              if (-1 * brett(n) * farbe >= OW) _
              and (-1 * brett(n) * farbe <= KW) then
                Eintragen(v, n)
              end if
            loop until brett(n) <> OW
          next i
        case KW
          for i = 1 to 8
            n = v + koenig(i)
            if (-1 * brett(n) * farbe >= OW) _
            and (-1 * brett(n) * farbe <= KW) then
              l = 0
              do
                l += 1
              loop until (-1 * brett(n + koenig(l)) * farbe = KW) or (l = 8)
              if l = 8 then Eintragen(v, n)
            end if
          next i
        case DW
          for i = 1 to 8
            n = v
            do
              n += koenig(i)
              if (-1 * brett(n) * farbe >= OW) _
              and (-1 * brett(n) * farbe <= KW) then
                Eintragen(v, n)
              end if
            loop until brett(n) <> OW
          next i
        end select
      end if
    next k
  next j
end sub

function tBrett.Schach as integer
  farbe = -1 * farbe
  ZuggeneratorOhne
  farbe = -1 * farbe
  for i as integer = 1 to anzahl
    if brett(liste.n(i)) = KW * farbe then
      return -1
    end if
  next i
  function = 0
end function

sub tBrett.RochadeTesten(vonkoenig as integer, schritt as integer, turmfeld as integer)
  dim b as tBrett
  dim i as integer
  dim rochade_verboten as integer = 0
  if brett(turmfeld) * farbe <> TW then exit sub
  i = vonkoenig + schritt
  do
    if brett(i) <> OW then exit sub
    i += schritt
  loop until i = turmfeld
  b = this
  b.farbe = -1 * b.farbe
  b.ZuggeneratorOhne
  for i = 1 to b.anzahl
    if (b.liste.n(i) mod 10 = vonkoenig mod 10) _
    andalso (b.liste.n(i) - vonkoenig) \ schritt >= 0 then
      rochade_verboten = -1
    end if
  next i
  if not rochade_verboten then Eintragen(vonkoenig, vonkoenig + 2 * schritt)
end sub

sub tBrett.Zuggenerator
  ZuggeneratorOhne
  if farbe = +1 then
    if instr(rochade, "K") > 0 then RochadeTesten(51, 10, 81)
    if instr(rochade, "Q") > 0 then RochadeTesten(51,-10, 11)
  end if
  if farbe = -1 then
    if instr(rochade, "k") > 0 then RochadeTesten(58, 10, 88)
    if instr(rochade, "q") > 0 then RochadeTesten(58,-10, 18)
  end if
end sub

function tBrett.RekursiveBewertung(f as integer, t as integer = 1, alfa as integer = 32000) as integer
  dim as integer w, i, beta, abbruch
  dim as tBrett b
  Zuggenerator
  i = 0
  beta = -32000 * farbe * f
  abbruch = 0
  while (i < anzahl) and not abbruch
    i += 1
    b = this
    b.Setzen(liste.v(i), liste.n(i))
    if ((t >= rechentiefe) and (brett(liste.n(i)) = OW)) or (t = maxtiefe) then
      w = b.Bewertung(f)
    else
      w = b.RekursiveBewertung(f, t + 1, beta)
    end if
    if farbe = f then
      if w > beta then beta = w
      if beta > alfa then abbruch = -1
    else
      if w < beta then beta = w
      if beta < alfa then abbruch = -1
    end if
    liste.w(i) = w
  wend
  function = beta
end function

data1:
data -010,-007,-006,-019,-126,-006,-007,-010
data -002,-002,-002,-002,-002,-002,-002,-002
data -000,-000,-000,-000,-000,-000,-000,-000
data -000,-000,-000,-000,-000,-000,-000,-000
data -000,-000,-000,-000,-000,-000,-000,-000
data -000,-000,-000,-000,-000,-000,-000,-000
data +002,+002,+002,+002,+002,+002,+002,+002
data +010,+007,+006,+019,+126,+006,+007,+010

sub tSchachprogramm.Grundstellung(s as string = "")
  dim as integer x, y, z
  with grundbrett
    for z = -10 to 109
      .brett(z) = AW
    next z
    restore data1
    for y = 8 to 1 step -1
      for x = 1 to 8
        read .brett(10 * x + y)
      next x
    next y
    .farbe = 1
    .rochade = "KQkq"
    .epfeld(+1) = 0
    .epfeld(-1) = 0
    .material = 0
  end with
  spielbrett = grundbrett
  halbzugzahl = 0
  if len(s) > 0 then
    dim as string symbol = "kqrbnp PNBRQK", s1 = s, a(1 to 6)
    z = instr(s1, " ") : a(1) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(2) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(3) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(4) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    z = instr(s1, " ") : a(5) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
    a(6) = s1
    with spielbrett
      x = 1
      y = 8
      for z = 1 to len(a(1))
        if a(1)[z - 1] = asc("/") then
          x = 1
          y -= 1
        end if
        if (a(1)[z - 1] >= asc("1")) andalso (a(1)[z - 1] <= asc("8")) then
          do while a(1)[z - 1] > asc("0")
            .brett(10 * x + y) = OW
            a(1)[z - 1] -= 1
            x += 1
          loop
        end if
        if instr(symbol, chr(a(1)[z - 1])) > 0 then
          select case instr(symbol, chr(a(1)[z - 1])) - 7
            case -6 : .brett(10 * x + y) = - 1 * KW
            case -5 : .brett(10 * x + y) = - 1 * DW
            case -4 : .brett(10 * x + y) = - 1 * TW
            case -3 : .brett(10 * x + y) = - 1 * LW
            case -2 : .brett(10 * x + y) = - 1 * SW
            case -1 : .brett(10 * x + y) = - 1 * BW
            case +1 : .brett(10 * x + y) = + 1 * BW
            case +2 : .brett(10 * x + y) = + 1 * SW
            case +3 : .brett(10 * x + y) = + 1 * LW
            case +4 : .brett(10 * x + y) = + 1 * TW
            case +5 : .brett(10 * x + y) = + 1 * DW
            case +6 : .brett(10 * x + y) = + 1 * KW
          end select
          x += 1
        end if
      next z
      .farbe = iif(a(2) = "w", +1, -1)
      .rochade = a(3)
      if len(a(4)) = 2 then
        .epfeld(-1 * .farbe) = 10 * (a(4)[0] - 96) + (a(4)[1] - 48)
      else
        .epfeld(-1 * .farbe) = 0
      end if
      .epfeld(.farbe) = 0
      for y = 8 to 1 step -1
        for x = 1 to 8
          .material += .brett(10 * x + y)
        next x
      next y
    end with
    halbzugzahl = 2 * (valint(a(6)) - 1) + iif(spielbrett.farbe = -1, 1, 0)
  end if
end sub

function tSchachprogramm.HtmlAnfang(font as string = "chess alfonso-x", size as string = "32px", colour as string = "midnightblue") as string
  function = _
  "<!DOCTYPE html>" & chr(13, 10) & _
  "<html>" & chr(13, 10) & _
  "<head>" & chr(13, 10) & _
  "  <title>JSSchach</title>" & chr(13, 10) & _
  "  <style type=""text/css"">" & chr(13, 10) & _
  "  p {" & chr(13, 10) & _
  "    font-family: " & font & ";" & chr(13, 10) & _
  "    font-size: " & size & ";" & chr(13, 10) & _
  "    color: " & colour & ";" & chr(13, 10) & _
  "  }" & chr(13, 10) & _
  "  </style>" & chr(13, 10) & _
  "</head>" & chr(13, 10) & _
  "<body>"
end function

function tSchachprogramm.Brettdarstellung as string
  dim as string char, s = _
  "<p>" & chr(13, 10) & _
  "!""""""""""""""""#<br>" & chr(13, 10) & _
  "$tMvWlVmT%<br>" & chr(13, 10) & _
  "$OoOoOoOo%<br>" & chr(13, 10) & _
  "$ + + + +%<br>" & chr(13, 10) & _
  "$+ + + + %<br>" & chr(13, 10) & _
  "$ + + + +%<br>" & chr(13, 10) & _
  "$+ + + + %<br>" & chr(13, 10) & _
  "$pPpPpPpP%<br>" & chr(13, 10) & _
  "$RnBqKbNr%<br>" & chr(13, 10) & _
  "/(((((((()<br>" & chr(13, 10) & _
  "</p>"
  for x as integer = 1 to 8
    for y as integer = 1 to 8
      select case spielbrett.brett(10 * x + y)
        case OW : char = " "
        case BW : char = "p"
        case SW : char = "n"
        case LW : char = "b"
        case TW : char = "r"
        case DW : char = "q"
        case KW : char = "k"
        case -1 * BW : char = "o"
        case -1 * SW : char = "m"
        case -1 * LW : char = "v"
        case -1 * TW : char = "t"
        case -1 * DW : char = "w"
        case -1 * KW : char = "l" 
      end select
      if (x + y) mod 2 = 0 then
        if spielbrett.brett(10 * x + y) = OW then
          char = "+"
        else
          char = chr(asc(char) - (asc("a") - asc("A")))
        end if
      end if
      s[x + 16 * (8 - y) + 22 - 1] = asc(char)
    next y
  next x
  function = s
end function

function tSchachprogramm.HtmlEnde as string
  function = _
  "</body>" & chr(13, 10) & _
  "</html>"
end function

function tSchachprogramm.Spielereingabe(s as string) as integer
  dim as integer result, zug, von, nach
  dim as tBrett b
  if (len(s) = 4) andalso _
  (s[0] >= asc("a")) and (s[0] <= asc("h")) and _
  (s[1] >= asc("1")) and (s[1] <= asc("8")) and _
  (s[2] >= asc("a")) and (s[2] <= asc("h")) and _
  (s[3] >= asc("1")) and (s[3] <= asc("8")) then
    s[0] -= (16 + 32)
    s[2] -= (16 + 32)
  else
    return 0
  end if
  result = 0
  zug = valint(s)
  spielbrett.Zuggenerator
  for i as integer = 1 to spielbrett.anzahl
    if zug = spielbrett.liste.v(i) * 100 + spielbrett.liste.n(i) then
      result = -1
      exit for
    end if
  next i
  if result = 0 then return 0
  von = zug \ 100
  nach = zug mod 100
  b = spielbrett
  b.Setzen(von, nach)
  b.farbe = -1 * b.farbe
  if b.Schach then
    result = 0
  else
    spielbrett.Setzen(von, nach)
    halbzugzahl += 1
  end if
  function = result
end function

function tSchachprogramm.Zugstring(von as integer, nach as integer) as string
  dim aux as string
  aux = str(von * 100 + nach)
  aux[0] += asc("a") - asc("1")
  aux[2] += asc("a") - asc("1")
  function = aux
end function

function tSchachprogramm.FeldIstInGrundstellung(byref b as tBrett, i as integer) as integer
  function = (b.brett(i) = grundbrett.brett(i))
end function

function tSchachprogramm.Besterzug(wert as integer) as integer
  dim b as tBrett
  dim as integer i, j, max
  anzahl2 = 0
  for i = 1 to spielbrett.anzahl
    if spielbrett.liste.w(i) = wert then
      anzahl2 += 1
      liste2.v(anzahl2) = spielbrett.liste.v(i)
      liste2.n(anzahl2) = spielbrett.liste.n(i)
    end if
  next i
  max = -1000
  function = 1
  for i = 1 to anzahl2
    liste2.w(i) = 0
    if FeldIstInGrundstellung(spielbrett, liste2.v(i)) then
      liste2.w(i) += 5
      if abs(spielbrett.brett(liste2.v(i))) = BW then
        liste2.w(i) += 2
      end if
    end if
    if abs(spielbrett.brett(liste2.v(i))) = KW then
      liste2.w(i) -= 10
    end if
    if (halbzugzahl < 32) and ( _
      (abs(spielbrett.brett(liste2.v(i))) = BW) or _
      (abs(spielbrett.brett(liste2.v(i))) = LW) or _
      (abs(spielbrett.brett(liste2.v(i))) = SW) _
    ) then
      liste2.w(i) += 20
    end if
    if (liste2.v(i) \ 10 = 1) _
    or (liste2.v(i) \ 10 = 8) _
    or (liste2.v(i) mod 10 = 1) _
    or (liste2.v(i) mod 10 = 8) then
      liste2.w(i) += 2
    end if
    if (liste2.n(i) \ 10 = 1) _
    or (liste2.n(i) \ 10 = 8) _
    or (liste2.n(i) mod 10 = 1) _
    or (liste2.n(i) mod 10 = 8) then
      liste2.w(i) -= 2
    end if
    b = spielbrett
    b.Setzen(liste2.v(i), liste2.n(i))
    if FeldIstInGrundstellung(b, liste2.n(i)) then
      liste2.w(i) -= 10
    end if
    b.farbe = -1 * b.farbe
    b.ZuggeneratorOhne
    for j = 1 to b.anzahl
      liste2.w(i) += 1
      if b.brett(b.liste.n(j)) <> OW then
        liste2.w(i) += 1
      end if
    next j
    b.farbe = -1 * b.farbe
    b.ZuggeneratorOhne
    for j = 1 to b.anzahl
      liste2.w(i) -= 1
      if b.brett(b.liste.n(j)) <> OW then
        liste2.w(i) -= 1
      end if
    next j
    if liste2.w(i) >= max then
      max = liste2.w(i)
      function = i
    end if
  next i
end function

function tSchachprogramm.Computerzug as string
  dim as integer i, s
  dim as string result, flagge
  i = Besterzug(spielbrett.RekursiveBewertung(spielbrett.farbe))
  halbzugzahl += 1
  s = spielbrett.Schach
  spielbrett.Setzen(liste2.v(i), liste2.n(i))
  result = Zugstring(liste2.v(i), liste2.n(i))
  spielbrett.farbe = -1 * spielbrett.farbe
  if spielbrett.Schach then
    if s then
      flagge = "3"
    else
      flagge = "2"
    end if
  else
    spielbrett.farbe = -1 * spielbrett.farbe
    if spielbrett.Schach then
      flagge = "1"
    else
      flagge = "0"
    end if
  end if
  result &= "_" & flagge
  function = result
end function
